package Lento::CacheTable;
use File::stat;
use XML::Simple;
use IO::File;
use strict;
use Lento::Debuggable;
use Lento::Fileset;
use Lento::Server;
use POE;

@Lento::CacheTable::ISA = qw(Lento::Debuggable);
$Lento::CacheTable::level = 1;

my $cachetable; 

sub new { 
    my $self = {};
    bless $self;

    $self->{fset2mtpt} = {};
    $self->{mtpt2fset} = {};
    return $self;
}

sub read_mounts {
    my $self = shift;
    my $stat;

    LOG ">>>>> Read InterMezzo Cache Table\n";
    my $ptr = $::psdev->readmount();
    my(@lines) = split(/\n/, $ptr);
    die "No InterMezzo file systems mounted yet" unless @lines;
    foreach my $line (@lines) {
	my ($fsetname, $path, $prestodev) = split (/  */, $line, 3);
	LOG ">>    Cache Table: $fsetname: $path, $prestodev\n";
	die "Protocol error" unless $prestodev eq $::prestodev;
	$stat = stat($path) || die "Cannot stat $path";
	$self->{dev2mtpt}->{$stat->dev} = $path; 
	$self->{fset2mtpt}->{$fsetname} = $path;
	$self->{mtpt2fset}->{$path} = $fsetname;
    }
    LOG "<<<<< Read InterMezzo Cache Table\n";
}

sub fset2cache { 
    my $self =  shift;
    my $fsetname = shift;
    return $self->{fset2mtpt}->{$fsetname};
}

sub printdb ($) {
    my $self = shift;
    my $result = "";
    my $name;

    foreach $name (keys %{$self->{fset2mtpt}}) {
	my $mtpt = $self->fset2cache($name);
	$result .= "fset name $name, mtpt: $mtpt\n";
    }
    return $result;
}


1;

__END__
