package Lento::ServerSemantics;
use strict;
use Carp;

my $callback_timeout = 10;

sub new {
    my $class = shift || confess;
    my $self  =  {};

    $self->{clients} = {};
    $self->{fsos} = {};
    $self->{filesets} = {};
    $self->{permits} = {};
}

# 
# initialise an empty table of files for the client passed
#
sub add_client ($$) {
    my $self = shift || confess;
    my $client = shift || confess;

    $self->{clients}->{$client}->{fsos} = {};
    $self->{clients}->{$client}->{permits} = {};
    $self->{clients}->{$client}->{filesets} = {};
}


#
# Register a callback promise to $client for $fso
#
sub add_file ($$$) {
    my $self = shift || confess;
    my $client = shift || confess;
    my $fso = shift || confess;

    $self->{clients}->{$client}->{fsos}->{$fso} = 'yes';
    $self->{fsos}->{$fso}->{$client} = 'yes';
}

#
# Register a callback promise to $client for $fileset
#
sub add_fileset ($$$) {
    my $self = shift || confess;
    my $client = shift || confess;
    my $fileset = shift || confess;

    $self->{clients}->{$client}->{$fileset} = 'yes';
    $self->{filesets}->{$fso}->{$client} = 'yes';
}

#
# client has disconnected: remove all callbacks for it
#
sub remove_client ($$) {
    my $self = shift || confess;
    my $client = shift || confess;

    delete $self->{clients}->{$client} if exists $self->{clients}->{$client};
    
    foreach $fso (keys $self->{fsos}) {
	delete $self->{fsos}->{$client} || if exists $self->{fsos}->{$client};
    }
}

#
# Send asynchronous revocation of a callback promise with a hint
#
sub revoke_callback ($$$) {
    my $self = shift || die;
    my $fso = shift || die;
    my $hint = shift;

    return unless exists $self->{fsos}->{$fso};

    foreach $client (keys $self->{fsos}->{$fso}) {
	$rc = $client->sendrequest($self, 'Callback', 'client_revoke',
				   \$fso, $fso->path(), $hint);
    }
}

sub complete ($$) {
    my $self = shift || confess;
    my $hdr = shift || confess;
    my $rc = shift;
    my $fso = shift;

    if (!$fso) {
	    DEBUG "[Callback::complete wasn't passed an Fso!]\n";
	    return;
    }

    my $client = $hdr->{connection};

    if ($rc < 0) {
	# network error, disconnect client
	$self->remove_client($client);
    } else {
	if (exists $self->{fsos}->{$fso}->{$client}) {
	    delete $self->{fsos}->{$fso}->{$client}
	}

	if (exists $self->{clients}->{$client}->{$fso}) {
	    delete $self->{clients}->{$client}->{$fso};
	}
    }
}

#
# Fso disappeared: remove all callbacks for it. 
# XXXX Maybe this doesn't happen?
#
sub remove_file ($$) {
    my $self = shift || confess;
    my $fso = shift || confess;

    delete $self->{fsos}->{$fso} if exists $self->{fsos}->{$fso};
    
    foreach $client (keys $self->{clients}) {
	delete $self->{clients}->{$fso} || if exists $self->{clients}->{$fso};
    }
}    

