package Lento::CreateML;

use strict;
use Lento::Debuggable;
use File::stat;
use POSIX qw(:sys_stat_h :errno_h :unistd_h :fcntl_h);
use POE;


sub __S_IFMT () {0170000;}
sub S_IFMT () { &__S_IFMT;}
sub __S_IFLNK () {0120000;}
sub S_IFLNK () { &__S_IFLNK;}

sub __S_ISTYPE {
    my ($mode, $mask) = @_;
    eval q(((($mode) &  &__S_IFMT) == ($mask)));
}
sub S_ISLNK {
    my $mode = shift;
    eval q( &__S_ISTYPE(($mode),  &__S_IFLNK));
}


$::Lento::CreateML::ISA = ("Debuggable");
$::Lento::CreateML::level = 1;

sub create_ml {
    my $path = shift;
    my $ml = {};
    my $recno = 0;
    my %inolist = ();
    `cd $path ;  find . -printf '/%P\n' > /tmp/ml.$$`;

    open FH, "/tmp/ml.$$";

    while ( <FH> ) {
	my @rec;
	chop;
	
	next if $_ eq '/';
	print("LSTAT $path : $_\n");
	my $st = lstat($path . "/" . $_) || die;

    if ($st->nlink > 1) {
    # potential hardlink
        if (exists($inolist{$st->ino})) {
        # some file with this inode number already exists...hard link to it!
           my $target=$inolist{$st->ino};
	       print("LINK: path $_, target: \"$target\"\n");
           @rec = (0, 0, $_, $inolist{$st->ino}, "LINK");
           $ml->{$recno++} = \@rec; 
           next;
        } else {
           $inolist{$st->ino}=$_;
        }
    }

	if ( S_ISREG($st->mode()) ) {
	    print("CREATE: path $_\n");
	    $ml->{$recno++} = [0, 0, $st->mode(), $_, "CREATE"];
	    $ml->{$recno++} = [0, 0, $st->uid(), $st->gid(), $st->size(), $st->atime(), $st->mtime(), $st->ctime(), $st->mode(), 0, ~0, $_, "SETATTR"];
	    $ml->{$recno++} = [0, 0, $_, "CLOSE"];
	}


	if ( S_ISDIR($st->mode()) ) {
	    print("MKDIR: path $_\n");
	    $ml->{$recno++} = [0, 0, $st->mode(), $_, "MKDIR"];
	    $ml->{$recno++} = [0, 0, $st->uid(), $st->gid(), $st->size(), $st->atime(), $st->mtime(), $st->ctime(), $st->mode(), 0, ~0, $_, "SETATTR"];
	}

	if ( S_ISLNK($st->mode()) ) {
	    my $target = readlink($path ."/" . $_);
	    print("SYM: path $_, target: \"$target\"\n");
	    @rec = (0, 0, $_, $target, "SYMLINK");
	    $ml->{$recno++} = \@rec;
	}

	if ( S_ISCHR($st->mode()) || S_ISBLK($st->mode()) ) {
	    my $dminor= 0xff & $st->rdev;
	    my $dmajor= (0xff00 & $st->rdev)>>8;

	    print("MKNOD %d, $dmajor, $dminor, $path\n", $st->mode());
	    $ml->{$recno++}= [0, 0, $st->mode, $dmajor, $dminor, $_, "MKNOD"];
            $ml->{$recno++} = [0, 0, $st->uid(), $st->gid(), $st->size(),
			       $st->atime(), $st->mtime(), $st->ctime(),
			       $st->mode(), 0, ~0, $_, "SETATTR"];
	}
    }
    close FH;

    unlink("/tmp/ml.$$");
    ($recno - 1, $ml);
}

