package Lento::Debuggable;
use POE;
use POE::Preprocessor;
use POE::Kernel;
use Data::Dumper;
use Exporter;
use POSIX qw(ENOSYS ctime);

@ISA = qw(Exporter);
@EXPORT = qw(DEBUG LOG DUMP TACITUS MarkLentoBusy);

$Lento::Debuggable::IDLE_LEVEL = -3;
$Lento::Debuggable::IDLE_GENERATION = 0;
# debug the sessions

# Borrowed from Poe::Kernel
# The Kernel object.  KR_SIZE goes last (it's the index count).
enum   KR_SESSIONS KR_VECTORS KR_HANDLES KR_STATES KR_SIGNALS KR_ALIASES
enum + KR_ACTIVE_SESSION KR_PROCESSES KR_ALARMS KR_ID KR_SESSION_IDS
enum + KR_ID_INDEX KR_WATCHER_TIMER KR_WATCHER_IDLE KR_EXTRA_REFS KR_SIZE

# State transition events.
enum ST_SESSION ST_SOURCE ST_NAME ST_TYPE ST_ARGS


sub debug_sessions {
    my $print_interval= shift;

    new POE::Session
        ( [$print_interval], 
          _start => sub {
              $_[HEAP]->{session_name} = $_[SESSION]->ID().":Session Debugger"; 
              $_[KERNEL]->alias_set("Session Debugger $_[SESSION]");
              $_[HEAP]->{delay}= $_[ARG0];
              $_[KERNEL]->yield('debug');
          },
          debug => sub {
              print "=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- Sessions\n";
              
              foreach my $session 
                  (keys %{$poe_kernel->[KR_SESSIONS]}) {
                      my $sess = $poe_kernel->alias_resolve($session);
                      my $name;
                      if (ref($sess) eq "POE::Session") {
                          my $heap = $sess->get_heap();

                          $name = defined $heap->{session_name} 
                                ? $sess->get_heap()->{session_name} : $session;
                          printf "=-=-=-> %-20s hash %s \n", $name, $session; 
                          if (index($name,"send_KML")>0 
                              ||index($name,"Reintegrate_KML")>0
                              ||index($name,"BackFetch")>0) {
                              &MarkLentoBusy;
                          }
                              
                          foreach $key (keys %{$heap}) {
                            printf "     -> $key size: " . 
                              length(Data::Dumper->Dump([$heap->{$key}])). "\n"; 
                          }

                      } else {
                          printf "=-=-=-> non session: %s\n", $session;
                      }
                  }
              
              print "\n=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- States\n";
              foreach my $enqueued_state
                  (@{$poe_kernel->[KR_STATES]}) {
                      my $session=$enqueued_state->[ST_SESSION];
                      my $state=$enqueued_state->[ST_NAME];
                      my $sess = $poe_kernel->alias_resolve($session);
                      my $name = $session;
                      if (ref($sess) eq "POE::Session") {
                          my $name = defined $session->get_heap()->{session_name} ? 
                              $session->get_heap()->{session_name} : $session;
                      } else {
                          my $name = $session;
                      }
                      printf "=-=-=-> %-20s hash %s state %s\n", $name, $session, $state;
                      &MarkLentoBusy;
                  }
              print "=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n";
              print "=-=-=IDLE_LEVEL: $Lento::Debuggable::IDLE_LEVEL $Lento::Debuggable::IDLE_GENERATION\n";
              if ($Lento::Debuggable::IDLE_LEVEL != -3){
                  $Lento::Debuggable::IDLE_LEVEL += 1;
              }
              $_[KERNEL]->delay('debug', $_[HEAP]->{delay});        
          }       
        );
}

#my loggingbits = 0xfffffff;

sub set_debug_level {
  my $class = shift;
  my $level = shift;
  ${"${class}::level"} = $level;   # this sets the level scalar symbol
  # in the package $class.
}

sub get_debug_level {
  my $class = shift;
  return ${"${class}::level"};
}

sub InterCon { 
    my $hdr = shift;
    push @Debuggable::intercons, $hdr;

#    $hdr->{connection}->sendreply($hdr->{token}, 0, undef);

}

sub MarkLentoBusy {
    $Lento::Debuggable::IDLE_GENERATION += 1
      if $Lento::Debuggable::IDLE_LEVEL > -1;
    $Lento::Debuggable::IDLE_LEVEL = -1;
}

sub LOG {
    my $msg = sprintf($_[0], @_[1..$#_]);
    my $time = substr(ctime(time()), 4, 15);

    print $time . ": " . $msg;
    foreach my $hdr ( @Debuggable::intercons ) {
            $hdr->{connection}->senddata($hdr->{token}, $msg);
    }
} 

sub DUMP {
    my ($class, $file, $line) = caller();
    my $level = "${class}::level";
    if ($$level + $::global_debug > 0) {
        print "[$file: $line] $_[0]:\n", Dumper(@_);
    
        return;
    }
}
    

sub DEBUG {
    my ($class, $file, $line) = caller();

    my $time = substr(ctime(time()), 11,8);

    my $level = "${class}::level";
    if (defined($$level) and defined ($::global_debug)) {
        if (${"${class}::level"} + $::global_debug > 0) {
            my @args =  @_;
                $args[0] = "[$file: $line: $time] $_[0]";
    
                printf @args;
    
                return;
            }
    }
}

sub set_option {
    my ($class, $file, $line) = caller();
    my $session = shift;
    my $level = "${class}::level";
    $session->option(@_) if ($$level + $::global_debug > 0);
}

sub TACITUS {
    my $str = shift;
    #TACITUS
    if (defined $::logfile) {
        open FH,">> ".$::logfile;
        print FH $str;
        close FH;
    }
}


sub test {
    $::global_debug =1;
    DEBUG "Hello world\n";
    LOG "info", "Hello world %s\n", "hoser";
}

1;

__END__
