# 
# Copyright (1999) Stelias Computing Inc, braam@stelias.com
# Copyright (1999) Carnegie Mellon University, braam@cs.cmu.edu
# 

package Lento::Filter::Packet;
use strict;
use Lento::ACT;
use Lento::Debuggable;
use Lento::Packet;
use Storable qw(nfreeze thaw);
#use MyPreProc;

@Lento::Filter::Packet::ISA = qw(Lento::Debuggable);
$Lento::Filter::Packet::level = -1000;

sub new {
  my $type = $_[0];
  return bless { basetype => $type }, $type;
}

sub put {
    my ($self, $packets) = @_;
    my @raw;
    foreach my $packet (@$packets) {
	    # print $packet->printme();
	    push(@raw, $packet->packme());
    }

    \@raw;
}

sub get {
    my ($self, $stream) = @_;
    my @result;
    my $left;
    my $offset = 0;

    $self->{buffer} .= join('', @$stream);
    #$self->{buffer} .= @{$stream}->[0];

    # adjust $left whenever we truncate the buffer
    $left = length($self->{buffer});
    my $bufsize = $left;

    while ( $left >= 0 ) {
	    DEBUG "GET: left $left, bufsize $bufsize\n";
	    if ( !$self->{packet} ) {
	        $self->{havehdr} = 0;
	        $self->{packet} = new Lento::Packet();
	    }

	    # have no header, want one
	    if ( $left >= HDRSIZE && !$self->{havehdr} ) {
	        $self->{packet}->getheader($self->{buffer}, $offset);
	        $self->{havehdr} = 1;
	        $offset += HDRSIZE;
	        $left -= HDRSIZE;
	        DEBUG "Adding header: left $left, bufsize $bufsize, bodysize $self->{packet}->[PKT_SIZE]\n";
	    }

	    # nothing useful can be grabbed
	    if ( !$self->{havehdr} ) {
	        DEBUG "No header: last, left $left, bufsize $bufsize, offset $offset\n";
	        last;
	    }
	
	    # here we know we have a header

	    # no body is following
	    if ($self->{packet}->[PKT_SIZE] == 0) {
	        DEBUG "No size: left $left, bufsize $bufsize\n";

	        push(@result, $self->{packet});
	        $self->{havehdr} = 0;
	        delete $self->{packet};
	        next;
	    } 

	    # we have a full body; add it to the result
	    if  ( $left >= $self->{packet}->[PKT_SIZE] ) {
	        DEBUG "--->left $left, bufsize $bufsize, offset $offset\n";
	        $self->{packet}->[PKT_BODY] = substr($self->{buffer}, $offset,
						    $self->{packet}->[PKT_SIZE]);
	        $offset += $self->{packet}->[PKT_SIZE];
	        $left -= $self->{packet}->[PKT_SIZE];
	        DEBUG "--->left $left, bufsize $bufsize, offset $offset\n";

	        DEBUG $self->{packet}->printme();
            #DEBUG $self->{packet}->[PKT_BODY];
	        # only REQ and REP need to have the body thawed
	        if ( $self->{packet}->[PKT_TYPE] == $Packet::REQ ||
	             $self->{packet}->[PKT_TYPE] == $Packet::REP ||
	             $self->{packet}->[PKT_TYPE] == $Packet::EOD ) {
		        my $refin = Storable::thaw $self->{packet}->[PKT_BODY];
		        $self->{packet}->[PKT_PARMS] = $refin;
                undef $self->{packet}->[PKT_BODY];
	        }

	        push(@result, $self->{packet});
	        $self->{havehdr} = 0;
	        delete $self->{packet};
	        next;
	    }
	    last;
    }

    # Truncate the buffer, discarding data already parsed
    $self->{buffer} = substr($self->{buffer}, $offset);

    # we will only trigger the receiver if a header is all we get
    \@result;
}

1;

__END__
