package Lento::InterMezzo::ReqHandler::InControl;
use strict;
use POE;
use Lento::Packet;
use Lento::Debuggable;

@Lento::InterMezzo::ReqHandler::InControl::ISA = qw(Lento::Debuggable);
$Lento::InterMezzo::ReqHandler::InControl::level = -1000;

sub Echo ($) {
    my $packet = shift;

    new POE::Session
      ([$packet, "InControl Echo"], # two arguments to _start
       'Lento::EventSupport', ['_stop', '_signal', '_child', 
							   'alias_set', 'alias_remove'],
       'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
       'REQ' => sub { 
	   my $pkt = $_[HEAP]->{req_packet}; # set by _start
	   my $line = shift(@{$pkt->[PKT_PARMS]});

	   $pkt->[PKT_CONNECTION]->sendpacket('REP', $pkt, [0, $line]);
	   $_[KERNEL]->yield('finish', [$pkt, 0]);
       }
      );
}

sub Fsetdb ($) {
    my $packet = shift;

    new POE::Session
      ([$packet, "InControl Fsetdb"], # two arguments to _start
       'Lento::EventSupport', ['_stop', '_signal', '_child',
							   'alias_set', 'alias_remove' ],
       'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
       'REQ' => sub { 
	   my $pkt = $_[HEAP]->{req_packet}; # set by _start
	   my $line = $::fsetdb->printdb();

	   $pkt->[PKT_CONNECTION]->sendpacket('REP', $pkt, [0, $line]);
	   $_[KERNEL]->yield('finish', [$pkt, 0]);
       }
      );
}

sub Debug ($) {
    my $packet = shift;

    new POE::Session
      ([$packet, "InControl Echo"], # two arguments to _start
       'Lento::EventSupport', ['_stop', '_signal', '_child',
							   'alias_set', 'alias_remove' ],
       'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
       'REQ' => sub { 
	   my $pkt = $_[HEAP]->{req_packet}; # set by _start
	   my $subsys = shift(@{$pkt->[PKT_PARMS]});
	   my $value = shift(@{$pkt->[PKT_PARMS]});

	   my $command = sprintf "\$%s::level = %d;", $subsys, $value;

	   DEBUG("Eval %s\n", $command);
	   my $rc = eval $command;

	   $pkt->[PKT_CONNECTION]->sendpacket('REP', $pkt, [$rc]);
	   $_[KERNEL]->yield('finish', [$pkt, 0]);
       }
      );
}

sub FlushPermit ($) {
    my $packet = shift;

    new POE::Session
      ([$packet, "InControl FlushPermit"], # two arguments to _start
       'Lento::EventSupport', ['_stop', '_signal', '_child',
							   'alias_set', 'alias_remove' ],
       'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
       'REQ' => sub { 
	   my $pkt = $_[HEAP]->{req_packet}; # set by _start
	   my $path = shift @{$pkt->[PKT_PARMS]};

	   $::psdev->revoke_permit($path);

	   $pkt->[PKT_CONNECTION]->sendpacket('REP', $pkt, [0]);
	   $_[KERNEL]->yield('finish', [$pkt, 0]);
       }
      );
}

sub FlushAttr ($) {
    my $packet = shift;

    new POE::Session
      ([$packet, "InControl FlushAttr"], # two arguments to _start
       'Lento::EventSupport', ['_stop', '_signal', '_child',
							   'alias_set', 'alias_remove' ],
       'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
       'REQ' => sub { 
	   my $pkt = $_[HEAP]->{req_packet}; # set by _start
	   my $path = shift @{$pkt->[PKT_PARMS]};

	   $::psdev->revoke_attr($path);

	   $pkt->[PKT_CONNECTION]->sendpacket('REP', $pkt, [0]);
	   $_[KERNEL]->yield('finish', [$pkt, 0]);
       }
      );
}

sub FlushData ($) {
    my $packet = shift;

    new POE::Session
      ([$packet, "InControl FlushData"], # two arguments to _start
       'Lento::EventSupport', ['_stop', '_signal', '_child',
							   'alias_set', 'alias_remove' ],
       'Lento::InterMezzo::ReqHandler', ['_start', 'finish', 'got_error'],
       'REQ' => sub { 
	   my $pkt = $_[HEAP]->{req_packet}; # set by _start
	   my $path = shift @{$pkt->[PKT_PARMS]};

	   $::psdev->revoke_data($path);

	   $pkt->[PKT_CONNECTION]->sendpacket('REP', $pkt, [0]);
	   $_[KERNEL]->yield('finish', [$pkt, 0]);
       }
      );
}

1;
