package Lento::InterMezzo::UpcallHandler;
use POE;
use Lento::RPCS::GetPermit;
use Lento::InterMezzo::ReqHandler;
use Lento::Connection;
use Lento::Debuggable;
use Lento::Packet;
use Lento::Filter::Upcall;
use Lento::ReqDispatcher;
use Lento::Server;
use Lento::EventSupport; 
use Lento::KML;
use Carp;
use POSIX qw(:errno_h :fcntl_h);

@Lento::Intermezzo::UpcallHandler::ISA = qw(Lento::Debuggable);
$Lento::InterMezzo::UpcallHandler::level = 1;

sub OpenFile ($) {
    my $upcall = shift;
    
    # XXX: Remove for DOD
    $upcall->complete(0);
    
    $upcall->{fset} = $::fsetdb->find_by_path($upcall->{path}) ||
    die "no fileset for " . $upcall->{path};
    
    if ($upcall->{fset}->serversysid() eq $::mysysid) {
    DEBUG "Not performing OpenFile for local fileset\n";
    $upcall->complete(0);
        return;
    }
    
    return;
}
          
sub OpenDir {
    my $upcall = shift;
    
    #XXX: Remove once fetch_on_demand is in place...
    LOG "OpenDir not yet implemented\n";
    $upcall->complete(0);
    return;
}

sub KML {
    my $upcall = shift;
    
    # we are given the 
    # - fsetname
    # - first and last record number
    # - offset of first record
    # - file size
    
    # TASK: send a segment to all replicators
    
    my $fset = $::fsetdb->find_by_name($upcall->{fileset});
    confess "No such fileset $upcall->{fileset}" unless $fset;
    DEBUG ">>>> Fileset: $upcall->{fileset} KML\n";
    
    $fset->{kml}->last_offset($upcall->{length});
    #   print $fset->{kml}->printme($upcall->{first_rec_offset}, "not brief");
    # This is a huge waste of resources for large KMLs
    #DEBUG $fset->{kml}->printme($upcall->{first_rec_offset}, "brief");
    $fset->{kml}->ReintegrateKML();
}

sub Cookie {
    my $upcall = shift;
    
    DEBUG "Processing cookie $upcall->{cookie}\n"; 
    Lento::EventSupport::RespondWithCookie($upcall);
}

#sub Open {
#}

sub Signal {
    
}

%::upcallspending= ();

sub Permit {
    my $upcall = shift;
    
    $upcall->{fset} = $::fsetdb->find_by_name($upcall->{fileset}) 
        || die "no fileset";
    DEBUG("  >>  UpcallHandler::Permit for $upcall->{path}, fsetname $upcall->{fset}->{name}, currently at $upcall->{fset}->{permitholder}\n");
    
    if ( $upcall->{fset}->{presto_read_only} ) {
        LOG("  >>  ERROR: Presto requests permit for read-only fileset %s\n",
        $upcall->{fset}->{name});
        $upcall->complete( EROFS );
    }
    
    new POE::Session 
       ( [$upcall],
            # Waited for by queued sessions
              'Lento::EventSupport', [ 'AddWaiter','alias_set', 'alias_remove',
                                     '_child', '_stop'], 
          _start => sub {
              my $upcall= $_[ARG0];
              $_[HEAP]->{session_name} = $_[SESSION]->ID().
              ": Permit UPC ($upcall->{path}) $_[SESSION]";
              $_[HEAP]->{fset}= $::fsetdb->find_by_name($upcall->{fileset}) 
                          or die "no fileset";
              my $fsetname=$_[HEAP]->{fset}->{name};
          
              Lento::Debuggable::set_option 
              ($_[SESSION], trace => 1, default => 1);
          
              my $pending=exists($::upcallspending{$fsetname});
          
              # Push into a list of pending upcalls for this fileset
              DEBUG "$_[HEAP]->{session_name} queuing upcall $upcall\n";
              push( @{$::upcallspending{$fsetname}}, $upcall); 
              DEBUG "upcall hashlist : @{[ keys %::upcallspending ]}\n";
          
              DEBUG "upcalls for $fsetname: ".@{$::upcallspending{$fsetname}}."\n";
          
              if ($pending) {
                  # Another session will take care of this upcall
                  DEBUG "upcalls already queued for this fileset returning\n";
                  return;
              }
          
              $_[KERNEL]->alias_set("Permit - $_[SESSION]");
              $_[KERNEL]->yield('_get_lock');
          },
          _get_lock => sub {
              # test sync lock
              if (defined $fset->{sync_session}) {
                  DEBUG "fileset $fsetname is syncing...waiting\n";
                  Lento::EventSupport::WaitForSession
                      ($fset->{sync_session}, 'get_lock');
                  return;
              }
          
              # test permit lock
              if (defined  $_[HEAP]->{fset}->{permit_session}) {
                  DEBUG "permit lock for $_[HEAP]->{fset}->{name}...waiting\n";
                  Lento::EventSupport::WaitForSession(
                              $_[HEAP]->{fset}->{permit_session}, '_get_lock');
                  return;
              } 
          
              # Add session to lock hash.
              DEBUG "got lock for $_[HEAP]->{fset}->{name}\n";
              $_[HEAP]->{fset}->{permit_session} = $_[SESSION];
          
              $_[KERNEL]->yield('_get_permit');
          },
          _get_permit => sub {
              if ( $_[HEAP]->{fset}->{permitholder} eq $::mysysid ) {
                  $_[KERNEL]->yield('_complete_upcalls',0);
                  return;
              }
          
              my $permitsession = Lento::RPCS::GetPermit::GetPermit
                  ($_[HEAP]->{fset}, '_complete_upcalls');
          },
          _complete_upcalls => sub {
              my $rc= $_[ARG0];
              my $upcall;

              DEBUG("<<<<  Permit UPC: done for ".$_[HEAP]->{fset}->{name}
              ."rc: $rc\n");
          
              if ($rc) {
                  # We currently use the pinger default delay to schedule
                  # a retry. 
                  # XXX: make this delay tunable!
                  my $delay=$Lento::Pinger::default_delay;
          
                  DEBUG "GetPermit returns error $rc, need to retry\n";
                  DEBUG "Unlocking, but will retry in $delay seconds!\n";
                  # Be a good samaritan and unlokc to let others fail as well!
                  # This will prevent normal permit sessions from hanging around
                  # and will also allow us to pool any additional upcalls during
                  # disconnection. Also it prevents priority inversion problems
                  # with sync sessions.
                  undef $_[HEAP]->{fset}->{permit_session};
                  $_[KERNEL]->delay('_get_lock',$delay);
                  return;
              }
              # Complete all pending upcalls 
              foreach $upcall (@{$::upcallspending{$_[HEAP]->{fset}->{name}}}) {
                  $upcall->complete($_[ARG0]);
                  DEBUG "completed upcall $upcall: ".$upcall->printme()."\n";
              }

              # Cleanup upcall list here so that future upcalls do not use 
              # this session for upcall completion
              delete ($::upcallspending{$_[HEAP]->{fset}->{name}});
              
              # Give presto time to process the upcalls
              $_[KERNEL]->yield('_release_lock');

              #XXX: very ugly way to force lento to give up the time slice
              sleep(1);

        },
        _release_lock => sub {
              DEBUG("Now releasing locks");
              
              $_[KERNEL]->alias_remove("Permit - $_[SESSION]");
              #unlock
              undef $_[HEAP]->{fset}->{permit_session};
              # Wake up all lock waiters
              Lento::EventSupport::WakeUpWaitList();
          }
       );
}

1;
