package Lento::KML::Rec;

# another new line

use strict;
use Lento::Debuggable;

@Lento::KML::Rec::ISA = qw(Lento::Debuggable);
$Lento::KML::Rec::level = 1;

my %opcode_unpack = (1 => \&unpack_create,
                     2 => \&unpack_mkdir,
                     3 => \&unpack_unlink,
                     4 => \&unpack_rmdir,
                     5 => \&unpack_close,
                     6 => \&unpack_symlink,
                     7 => \&unpack_rename,
                     8 => \&unpack_setattr,
                     9 => \&unpack_link,
                    10 => \&unpack_open,
                    11 => \&unpack_mknod,
                    15 => \&unpack_extattr,
                    16 => \&unpack_extattr);

sub size_round {
    my $val = shift;
    return ($val +3) & (~0x3);
}

my $KML_PREFIX_WORDS= 8;

sub Unpack {
    my $class = shift;
    my $self = bless {};
    my $buffer = shift;
    my $offset = shift;
    my $kml_offset = shift;
    my $length  = length($buffer) - $offset;

    die unless $length > 0;
    $offset = 0 unless defined $offset;

    $self->{kml_offset} = $kml_offset;
    ($self->{len}, $self->{version}, $self->{pid}, $self->{auid}, 
     $self->{fsuid}, $self->{fsgid}, $self->{opcode}, $self->{ngroups}) 
        = unpack("x$offset i$KML_PREFIX_WORDS", $buffer);

    $self->{groups} = [];
    my $gr_offset = $offset + $KML_PREFIX_WORDS * 4; 
    @{$self->{groups}} = unpack("x$gr_offset i$self->{ngroups}", $buffer); 

    my $func = $opcode_unpack{$self->{opcode}};
    die "Unknown opcode $self->{opcode} at offset $offset" unless $func;
    
    &{$func}($self, $buffer, $offset + $KML_PREFIX_WORDS * 4 + 
            $self->{ngroups} * 4);

    my $len = $self->{offs};
    ($self->{prevrec}, $self->{recno}, $self->{time}, $self->{slen}) = 
        unpack("x$len IIII", $buffer);

    # if the length stored in the prefix is not the same as the 
    #   length stored in the suffix then we have a serious problem.
    # We will return undef, and let the caller decide what to do.
    if ($self->{len} != $self->{slen}) {
        DEBUG "len $self->{len}, pid $self->{pid}, auid $self->{auid},". 
          "fsuid $self->{fsuid}, fsgid $self->{fsgid},".
          "opcode $self->{opcode}, prevrec $self->{prevrec},".
          "recno $self->{recno}, time $self->{time}, slen $self->{slen}\n";

        DEBUG $self->printme();
        die "corrupted KML? KML Unpack failed: $self->{len} != $self->{slen}\n";
    }

    $self;
}

sub printme {
    my $self = shift;
    my $format = shift;
    my $str;

    #TACITUS         
    my $fset=shift;   
    my $fsetpath="";  
    $fsetpath= $fset->mtpt() if defined $fset;

    if (defined $format and $format eq "brief") {
        $str .= sprintf " %08d", $self->{recno} if defined $self->{recno};
        $str .= sprintf " %7s", $self->{op} if defined $self->{op};
        $str .= sprintf " $self->{path}" if defined $self->{path};
        $str .= sprintf " $self->{target}" if defined $self->{target};
        $str .= "\n";
        return $str;
    }

    # XXX shirish: removed $::logfile below!
    #TACITUS
    if (defined $format and $format eq "verybrief") {
        $str .= sprintf " %7s", $self->{op} if defined $self->{op};
        if ($self->{opcode}==3 || $self->{opcode}==4) {
         #$str .= sprintf " ".$fsetpath.$self->{path} if defined $self->{path};
         if (defined $self->{target}) {
             $str .= " ";
             $str .= $fsetpath."/" if !($self->{target} =~ m,^/,);
             $str .= $self->{target};
         }
        } else {
            if (defined $self->{path}) {
                $str .= " ";
                $str .= $fsetpath.$self->{path};
            }
            if (defined $self->{target}) {
                $str .= " ";
                $str .= $fsetpath."/" if !($self->{target} =~ m,^/,);
                $str .= $self->{target};
            }
        }
        # XXX removed this since it produces an error message. 
        #$str .= sprintf " ".getpwuid($self->{uid})
        #    if (defined $self->{uid} && $self->{uid});
        # XXX clean up 
        #$str .= sprintf ":".getgrgid($self->{gid}) 
        #if defined $self->{gid} && $self->{gid};
        $str .= "\n";
        return $str;
    }


    $str = " -- Record:\n";
    $str .= sprintf "    Recno     $self->{recno}\n" if defined $self->{recno};
    $str .= sprintf "    KML off   $self->{kml_offset}\n" 
                            if defined $self->{kml_offset};
    $str .= sprintf "    Version   $self->{version}\n" 
                            if defined $self->{version};
    $str .= sprintf "    Len       $self->{len}\n" if defined $self->{len} ;
    $str .= sprintf "    Suf len   $self->{slen}\n" if defined $self->{slen} ;
    $str .= sprintf "    Time      $self->{time}\n" if defined $self->{time} ;
    $str .= sprintf "    Opcode    $self->{opcode}\n" 
        if defined $self->{opcode} ;
    $str .= sprintf "    Op        $self->{op}\n" if defined $self->{op};
    $str .= sprintf "    Pid       $self->{pid}\n" if defined $self->{pid} ;
    $str .= sprintf "    AUid      $self->{auid}\n" if defined $self->{auid} ;
    $str .= sprintf "    Fsuid     $self->{fsuid}\n" 
        if defined $self->{fsuid} ;
    $str .= sprintf "    Fsgid     $self->{fsgid}\n" 
        if defined $self->{fsgid} ;
    $str .= sprintf "    Prevrec   $self->{prevrec}\n" 
        if defined $self->{prevrec} ;
    $str .= sprintf "    Ngroups   $self->{ngroups}\n" 
        if defined $self->{groups} ;
    $str .= sprintf "    Groups    @{$self->{groups}}\n" 
        if defined $self->{groups} ;
    $str .= " -- Path:\n";
    $str .= sprintf "    Inode     $self->{ino}\n" if defined $self->{ino};
    $str .= sprintf("    Gen num   %u\n", $self->{generation}) 
        if defined $self->{generation} ;
    $str .= sprintf "    Path      $self->{path}\n" if defined $self->{path};
    $str .= sprintf("    Open_mode %o\n", $self->{open_mode}) if defined $self->{open_mode};
    $str .= sprintf "    Open_uid $self->{open_uid}\n" if defined $self->{open_uid};
    $str .= sprintf "    Open_gid $self->{open_gid}\n" if defined $self->{open_gid};
    $str .= sprintf "    Pathlen   $self->{pathlen}\n" 
        if defined $self->{pathlen} ;
    $str .= sprintf "    Tgt       $self->{target}\n" if defined $self->{target} ;
    $str .= sprintf "    Tgtlen    $self->{targetlen}\n" 
        if defined $self->{targetlen} ;

    $str .= " -- Attr:\n";
    $str .= sprintf("    Valid     %x\n", $self->{valid}) 
        if defined $self->{valid};
    $str .= sprintf("    Mode      %o\n", $self->{mode})
        if defined $self->{mode};
    $str .= sprintf "    Gid       $self->{gid}\n" if defined $self->{gid};
    $str .= sprintf "    Uid       $self->{uid}\n" if defined $self->{uid};
    $str .= sprintf "    Size      $self->{size}\n" if defined $self->{size};
    $str .= sprintf "    mtime     $self->{mtime}\n" if defined $self->{mtime};
    $str .= sprintf "    ctime     $self->{ctime}\n" if defined $self->{ctime};
    $str .= sprintf("    flags   %x\n", $self->{flags}) 
        if defined $self->{flags};
    $str .= sprintf "    Major     $self->{major}\n" if defined $self->{major};
    $str .= sprintf "    Minor     $self->{minor}\n" if defined $self->{minor};
    $str .= sprintf "    EA Flags  $self->{attr_flags}\n" 
        if defined $self->{attr_flags};
    $str .= sprintf "    EA Namlen $self->{namelen}\n" 
        if defined $self->{namelen};
    $str .= sprintf "    EA Name   $self->{attr_name}\n" 
        if defined $self->{attr_name};
    # do this to detect null value attributes
    $str .= sprintf "    EA Buflen $self->{buflen}\n" 
        if defined $self->{buflen};
    $str .= sprintf "    EA Value  $self->{attr_value}\n" 
        if defined $self->{attr_value};

    $str .= " -- Version:\n";
    # object information
    $str .= sprintf "    New mtime  $self->{new_obj_mtime}\n" 
        if defined $self->{new_obj_mtime};
    $str .= sprintf "    New ctime  $self->{new_obj_ctime}\n" 
        if defined $self->{new_obj_ctime};
    $str .= sprintf "    New size   $self->{new_obj_size}\n" 
        if defined $self->{new_obj_size};
    $str .= sprintf "    Old mtime  $self->{old_obj_mtime}\n" 
        if defined $self->{old_obj_mtime};
    $str .= sprintf "    Old ctime  $self->{old_obj_ctime}\n" 
        if defined $self->{old_obj_ctime};
    $str .= sprintf "    Old size   $self->{old_obj_size}\n" 
        if defined $self->{old_obj_size};
    # target information
    $str .= sprintf "    Old tgt mt $self->{old_tgt_mtime}\n" 
        if defined $self->{old_tgt_mtime};
    $str .= sprintf "    Old tgt ct $self->{old_tgt_ctime}\n" 
        if defined $self->{old_tgt_ctime};
    $str .= sprintf "    Old tgt sz $self->{old_tgt_size}\n" 
        if defined $self->{old_tgt_size};
    $str .= sprintf "    New tgt mt $self->{new_tgt_mtime}\n" 
        if defined $self->{new_tgt_mtime};
    $str .= sprintf "    New tgt ct $self->{new_tgt_ctime}\n" 
        if defined $self->{new_tgt_ctime};
    $str .= sprintf "    New tgt sz $self->{new_tgt_size}\n" 
        if defined $self->{new_tgt_size};

    # parent directory information
    $str .= sprintf "    Old p mt   $self->{old_pdir_mtime}\n" 
        if defined $self->{old_pdir_mtime};
    $str .= sprintf "    Old p ct   $self->{old_pdir_ctime}\n" 
        if defined $self->{old_pdir_ctime};
    $str .= sprintf "    Old p sz   $self->{old_pdir_size}\n" 
        if defined $self->{old_pdir_size};
    $str .= sprintf "    New p mt   $self->{new_pdir_mtime}\n" 
        if defined $self->{new_pdir_mtime};
    $str .= sprintf "    New p ct   $self->{new_pdir_ctime}\n" 
        if defined $self->{new_pdir_ctime};
    $str .= sprintf "    New p sz   $self->{new_pdir_size}\n" 
        if defined $self->{new_pdir_size};


    $str;
}

sub unpack_create {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 88;

    DEBUG("POS $pos\n");
    ($self->{old_pdir_mtime}, $self->{old_pdir_ctime}, $self->{old_pdir_size}, 
     $self->{new_pdir_mtime}, $self->{new_pdir_ctime}, $self->{new_pdir_size}, 
     $self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{mode}, $self->{uid}, $self->{gid}, $self->{pathlen}) = 
         unpack("x$pos ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 iiii", $buf);

    $self->{op} = "CREATE";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen});

    return $self;
}

sub unpack_symlink ($$) {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 88;

    DEBUG("POS $pos\n");
    ($self->{old_pdir_mtime}, $self->{old_pdir_ctime}, $self->{old_pdir_size}, 
     $self->{new_pdir_mtime}, $self->{new_pdir_ctime}, $self->{new_pdir_size}, 
     $self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{uid}, $self->{gid}, $self->{pathlen}, $self->{targetlen}) = 
         unpack("x$pos ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 iiii", $buf);

    $self->{op} = "SYMLINK";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{target} = substr($buf, $pos + $unpack_size + 
                             size_round($self->{pathlen}), $self->{targetlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen}) + 
        size_round($self->{targetlen});

    return $self;
}

sub unpack_mkdir {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 88;

    DEBUG("POS $pos\n");
    ($self->{old_pdir_mtime}, $self->{old_pdir_ctime}, $self->{old_pdir_size}, 
     $self->{new_pdir_mtime}, $self->{new_pdir_ctime}, $self->{new_pdir_size}, 
     $self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{mode}, $self->{uid}, $self->{gid}, $self->{pathlen}) = 
         unpack("x$pos ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 iiii", $buf);

    $self->{op} = "MKDIR";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen});

    return $self;
}


sub unpack_mknod {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 96;

    DEBUG("POS $pos\n");
    ($self->{old_pdir_mtime}, $self->{old_pdir_ctime}, $self->{old_pdir_size}, 
     $self->{new_pdir_mtime}, $self->{new_pdir_ctime}, $self->{new_pdir_size}, 
     $self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{mode}, $self->{uid}, $self->{gid}, 
     $self->{major}, $self->{minor}, $self->{pathlen}) = 
         unpack("x$pos ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 iiiiii", $buf);

    $self->{op} = "MKNOD";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen});

    return $self;
}

sub unpack_link {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 80;

    DEBUG("POS $pos\n");
    ($self->{old_pdir_mtime}, $self->{old_pdir_ctime}, $self->{old_pdir_size}, 
     $self->{new_pdir_mtime}, $self->{new_pdir_ctime}, $self->{new_pdir_size}, 
     $self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{pathlen}, $self->{targetlen}) = 
         unpack("x$pos ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 ii", $buf);

    $self->{op} = "LINK";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{target} = substr($buf, $pos + $unpack_size + 
                             size_round($self->{pathlen}), $self->{targetlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen}) + 
        size_round($self->{targetlen});

    return $self;
}

sub unpack_rename {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 104;

    # special case - names don't match reality:
    #     old_obj: src parent dir old times
    #     new_obj: src parent dir new times
    #     old_tgt: tgt parent dir old times
    #     new_tgt: tgt parent dir new times
    # this is because the times in the object itself don't change
    # but it's old parent dir (src parent dir) and it's new parent dir
    # (tgt parent dir) times are updated.
    #
    DEBUG("POS $pos\n");
    ($self->{old_obj_mtime}, $self->{old_obj_ctime}, $self->{old_obj_size}, 
     $self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{old_tgt_mtime}, $self->{old_tgt_ctime}, $self->{old_tgt_size}, 
     $self->{new_tgt_mtime}, $self->{new_tgt_ctime}, $self->{new_tgt_size}, 
     $self->{pathlen}, $self->{targetlen}) = 
         unpack("x$pos ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 ii", $buf);

    $self->{op} = "RENAME";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{target} = substr($buf, $pos + $unpack_size + 
                             size_round($self->{pathlen}), $self->{targetlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen}) + 
        size_round($self->{targetlen});;

    return $self;
}

sub unpack_unlink {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 80;

    DEBUG("POS $pos\n");
    ($self->{old_pdir_mtime}, $self->{old_pdir_ctime}, $self->{old_pdir_size}, 
     $self->{new_pdir_mtime}, $self->{new_pdir_ctime}, $self->{new_pdir_size}, 
     $self->{old_tgt_mtime}, $self->{old_tgt_ctime}, $self->{old_tgt_size}, 
     $self->{pathlen}, $self->{targetlen}) = 
         unpack("x$pos ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 ii", $buf);

    $self->{op} = "UNLINK";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{target} = substr($buf, $pos + $unpack_size + 
                             size_round($self->{pathlen}), $self->{targetlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen}) + 
        size_round($self->{targetlen});

    return $self;
}

sub unpack_rmdir {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 80;

    DEBUG("POS $pos\n");
    ($self->{old_pdir_mtime}, $self->{old_pdir_ctime}, $self->{old_pdir_size}, 
     $self->{new_pdir_mtime}, $self->{new_pdir_ctime}, $self->{new_pdir_size}, 
     $self->{old_tgt_mtime}, $self->{old_tgt_ctime}, $self->{old_tgt_size}, 
     $self->{pathlen}, $self->{targetlen}) = 
         unpack("x$pos ix4ix4ix4 ix4ix4ix4 ix4ix4ix4 ii", $buf);

    $self->{op} = "RMDIR";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{target} = substr($buf, $pos + $unpack_size + 
                             size_round($self->{pathlen}), $self->{targetlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen}) + 
        size_round($self->{targetlen});

    return $self;
}

sub unpack_setattr {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 72;

    DEBUG("POS $pos\n");
    ($self->{old_obj_mtime}, $self->{old_obj_ctime}, $self->{old_obj_size},  
     $self->{valid}, $self->{mode}, $self->{uid}, $self->{gid}, 
     $self->{size}, $self->{mtime}, $self->{ctime},  
     $self->{flags}, $self->{pathlen}) = 
         unpack("x$pos ix4ix4ix4 iiii ix4ix4ix4 ii", $buf);

    $self->{op} = "SETATTR";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen});

    return $self;
}

sub unpack_close {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 52;

    DEBUG("POS $pos\n");
    ($self->{open_mode}, $self->{open_uid}, $self->{open_gid}, 
     $self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{ino}, $self->{generation}, $self->{pathlen}) = 
         unpack("x$pos iiiix4ix4ix4 ix4ii", $buf);

    $self->{op} = "CLOSE";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen});

    return $self;
}


# Unpack extended attributes
sub unpack_extattr {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 68;

    DEBUG("POS $pos\n");
    ($self->{old_obj_mtime}, $self->{old_obj_ctime}, $self->{old_obj_size}, 
     $self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{attr_flags}, $self->{mode}, $self->{pathlen}, $self->{namelen}, 
     $self->{buflen}) = unpack("x$pos ix4ix4ix4 ix4ix4ix4 iiiii", $buf);

    #XXX: Ugh...lazy shirish, using raw opcodes!!! write a new function? 
    $self->{op} = ($self->{opcode} == 15) ? "SETEXTATTR" : "DELEXTATTR";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{attr_name} = substr($buf, $pos + $unpack_size 
                               +size_round($self->{pathlen}), 
                               $self->{namelen});

    # only unpack value if this is not a delete 
    if ($self->{opcode} == 15) {
        # Note that buflen can be zero for null attribute values 
        $self->{attr_value} = substr($buf, $pos + $unpack_size 
                                     +size_round($self->{pathlen})
                                     +size_round($self->{namelen}), 
                                     $self->{buflen});
    } else {
        $self->{buflen} = 0;
    }

    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen}) + 
                    size_round($self->{namelen}) + size_round($self->{buflen});

    return $self;
}


1;

__END__
