package Lento::LML::Rec;

# another new line

use strict;
use Lento::Debuggable;

@Lento::LML::Rec::ISA = qw(Lento::Debuggable);
$Lento::LML::Rec::level = 1;

# We still do KML style unpacking because LML might get extended to other
# types of operations.
my %opcode_unpack = ( 0 => \&unpack_noop,
					  5 => \&unpack_close );

sub size_round {
    my $val = shift;
    return ($val +3) & (~0x3);
}

my $LML_PREFIX_WORDS= 8;

sub Unpack {
    my $class = shift;
    my $self = bless {};
    my $buffer = shift;
    my $offset = shift;
    my $LML_offset = shift;
    my $length  = length($buffer) - $offset;

    die unless $length > 0;
    $offset = 0 unless defined $offset;

    $self->{LML_offset} = $LML_offset;
    ($self->{len}, $self->{version}, $self->{pid}, $self->{auid}, 
     $self->{fsuid}, $self->{fsgid}, $self->{opcode}, $self->{ngroups}) 
		= unpack("x$offset i$LML_PREFIX_WORDS", $buffer);

    $self->{groups} = [];
    my $gr_offset = $offset + $LML_PREFIX_WORDS * 4; 
    @{$self->{groups}} = unpack("x$gr_offset i$self->{ngroups}", $buffer); 

    if ($self->{opcode} == 0) {
		DEBUG "NOOP in LML\n";
        return undef;
	}

    my $func = $opcode_unpack{$self->{opcode}};
    die "Unknown opcode $self->{opcode} at offset $offset" unless $func;
    
    &{$func}($self, $buffer, $offset + $LML_PREFIX_WORDS * 4 + 
	    $self->{ngroups} * 4);

    my $len = $self->{offs};
    ($self->{prevrec}, $self->{recno}, $self->{time}, $self->{slen}) = 
	unpack("x$len IIII", $buffer);

    # if the length stored in the prefix is not the same as the 
    #   length stored in the suffix then we have a serious problem.
    # We will return undef, and let the caller decide what to do.
    if ($self->{len} != $self->{slen}) {
        DEBUG "len $self->{len}, pid $self->{pid}, auid $self->{auid},". 
          "fsuid $self->{fsuid}, fsgid $self->{fsgid},".
          "opcode $self->{opcode}, prevrec $self->{prevrec},".
          "recno $self->{recno}, time $self->{time}, slen $self->{slen}\n";

        DEBUG $self->printme();
        die "Corrupted LML? LML Unpack failed: $self->{len} != $self->{slen}\n";
    }

    $self;
}

sub printme {
    my $self = shift;
    my $format = shift;
    my $str;

    #TACITUS         
    my $fset=shift;   
    my $fsetpath="";  
    $fsetpath= $fset->mtpt() if defined $fset;

    if (defined $format and $format eq "brief") {
	$str .= sprintf " %08d", $self->{recno} if defined $self->{recno};
	$str .= sprintf " %7s", $self->{op} if defined $self->{op};
	$str .= sprintf " $self->{path}" if defined $self->{path};
	$str .= sprintf " $self->{target}" if defined $self->{target};
	$str .= "\n";
	return $str;
    }

    # XXX shirish: removed $::logfile below!
    #TACITUS
    if (defined $format and $format eq "verybrief") {
        $str .= sprintf " %7s", $self->{op} if defined $self->{op};
        if ($self->{opcode}==3 || $self->{opcode}==4) {
         #$str .= sprintf " ".$fsetpath.$self->{path} if defined $self->{path};
         if (defined $self->{target}) {
             $str .= " ";
             $str .= $fsetpath."/" if !($self->{target} =~ m,^/,);
             $str .= $self->{target};
         }
        } else {
            if (defined $self->{path}) {
                $str .= " ";
                $str .= $fsetpath.$self->{path};
            }
            if (defined $self->{target}) {
                $str .= " ";
                $str .= $fsetpath."/" if !($self->{target} =~ m,^/,);
                $str .= $self->{target};
            }
        }
        $str .= sprintf " ".getpwuid($self->{uid})
            if (defined $self->{uid} && $self->{uid});
	# XXX clean up 
        #$str .= sprintf ":".getgrgid($self->{gid}) 
        #if defined $self->{gid} && $self->{gid};
        $str .= "\n";
        return $str;
    }


    $str = " -- Record:\n";
    $str .= sprintf "    Recno     $self->{recno}\n" if defined $self->{recno};
    #$str .= sprintf "    LML off   $self->{LML_offset}\n" 
	#                    if defined $self->{LML_offset};
    $str .= sprintf "    lml off   $self->{lml_offset}\n" 
	                    if defined $self->{lml_offset};
    $str .= sprintf "    Len       $self->{len}\n" if defined $self->{len} ;
    $str .= sprintf "    Suf len   $self->{slen}\n" if defined $self->{slen} ;
    $str .= sprintf "    Time      $self->{time}\n" if defined $self->{time} ;
    $str .= sprintf "    Opcode    $self->{opcode}\n" 
	if defined $self->{opcode} ;
    $str .= sprintf "    Op        $self->{op}\n" if defined $self->{op};
    $str .= sprintf "    Pid       $self->{pid}\n" if defined $self->{pid} ;
    $str .= sprintf "    AUid      $self->{auid}\n" if defined $self->{auid} ;
    $str .= sprintf "    Fsuid     $self->{fsuid}\n" 
	if defined $self->{fsuid} ;
    $str .= sprintf "    Fsgid     $self->{fsgid}\n" 
	if defined $self->{fsgid} ;
    $str .= sprintf "    Prevrec   $self->{prevrec}\n" 
	if defined $self->{prevrec} ;
    $str .= sprintf "    Ngroups   $self->{ngroups}\n" 
	if defined $self->{groups} ;
    $str .= sprintf "    Groups    @{$self->{groups}}\n" 
	if defined $self->{groups} ;
    $str .= " -- Path:\n";
    $str .= sprintf "    Inode     $self->{ino}\n" if defined $self->{ino};
    $str .= sprintf "    RIno      $self->{remote_ino}\n" if defined $self->{remote_ino};
    $str .= sprintf("    Gen num   %u\n", $self->{generation}) 
	if defined $self->{generation} ;
    $str .= sprintf "    Path      $self->{path}\n" if defined $self->{path};
    $str .= sprintf "    Pathlen   $self->{pathlen}\n" 
	if defined $self->{pathlen} ;
    $str .= sprintf "    Tgt       $self->{target}\n" if defined $self->{target} ;
    $str .= sprintf "    Tgtlen    $self->{targetlen}\n" 
	if defined $self->{targetlen} ;

    $str .= " -- Attr:\n";
    $str .= sprintf("    Valid     %x\n", $self->{valid}) 
	if defined $self->{valid};
    $str .= sprintf("    Mode      %o\n", $self->{mode})
	if defined $self->{mode};
    $str .= sprintf "    Gid       $self->{gid}\n" if defined $self->{gid};
    $str .= sprintf "    Uid       $self->{uid}\n" if defined $self->{uid};
    $str .= sprintf "    Size      $self->{size}\n" if defined $self->{size};
    $str .= sprintf "    mtime     $self->{mtime}\n" if defined $self->{mtime};
    $str .= sprintf "    ctime     $self->{ctime}\n" if defined $self->{ctime};
    $str .= sprintf("    flags   %x\n", $self->{flags}) 
	if defined $self->{flags};
    $str .= sprintf "    Major     $self->{major}\n" if defined $self->{major};
    $str .= sprintf "    Minor     $self->{minor}\n" if defined $self->{minor};

    $str .= " -- Version:\n";
    # object information
    $str .= sprintf "    New mtime  $self->{new_obj_mtime}\n" 
	if defined $self->{new_obj_mtime};
    $str .= sprintf "    New ctime  $self->{new_obj_ctime}\n" 
	if defined $self->{new_obj_ctime};
    $str .= sprintf "    New size   $self->{new_obj_size}\n" 
	if defined $self->{new_obj_size};
    $str .= sprintf "    Old mtime  $self->{old_obj_mtime}\n" 
	if defined $self->{old_obj_mtime};
    $str .= sprintf "    Old ctime  $self->{old_obj_ctime}\n" 
	if defined $self->{old_obj_ctime};
    $str .= sprintf "    Old size   $self->{old_obj_size}\n" 
	if defined $self->{old_obj_size};
    # target information
    $str .= sprintf "    Old tgt mt $self->{old_tgt_mtime}\n" 
	if defined $self->{old_tgt_mtime};
    $str .= sprintf "    Old tgt ct $self->{old_tgt_ctime}\n" 
	if defined $self->{old_tgt_ctime};
    $str .= sprintf "    Old tgt sz $self->{old_tgt_size}\n" 
	if defined $self->{old_tgt_size};
    $str .= sprintf "    New tgt mt $self->{new_tgt_mtime}\n" 
	if defined $self->{new_tgt_mtime};
    $str .= sprintf "    New tgt ct $self->{new_tgt_ctime}\n" 
	if defined $self->{new_tgt_ctime};
    $str .= sprintf "    New tgt sz $self->{new_tgt_size}\n" 
	if defined $self->{new_tgt_size};

    # parent directory information
    $str .= sprintf "    Old p mt   $self->{old_pdir_mtime}\n" 
	if defined $self->{old_pdir_mtime};
    $str .= sprintf "    Old p ct   $self->{old_pdir_ctime}\n" 
	if defined $self->{old_pdir_ctime};
    $str .= sprintf "    Old p sz   $self->{old_pdir_size}\n" 
	if defined $self->{old_pdir_size};
    $str .= sprintf "    New p mt   $self->{new_pdir_mtime}\n" 
	if defined $self->{new_pdir_mtime};
    $str .= sprintf "    New p ct   $self->{new_pdir_ctime}\n" 
	if defined $self->{new_pdir_ctime};
    $str .= sprintf "    New p sz   $self->{new_pdir_size}\n" 
	if defined $self->{new_pdir_size};


    $str;
}

sub unpack_close {
    my ($self, $buf, $pos) = @_;
    my $unpack_size = 64;

    DEBUG("POS $pos\n");
    ($self->{new_obj_mtime}, $self->{new_obj_ctime}, $self->{new_obj_size}, 
     $self->{ino}, $self->{generation}, $self->{pathlen}, 
     $self->{remote_ino}, $self->{remote_generation}, $self->{remote_version},
     $self->{lml_offset}) = 
	 unpack("x$pos ix4ix4ix4 ix4ii ix4ii ix4", $buf);

    print "mismatch in lml offset" 
        if $self->{lml_offset} != $self->{LML_offset};

    $self->{op} = "CLOSE";
    $self->{path} = substr($buf, $pos + $unpack_size, $self->{pathlen});
    $self->{offs} = $pos + $unpack_size + size_round($self->{pathlen});

    return $self;
}


1;

__END__
