# 
# Copyright (1999) Stelias Computing Inc, braam@stelias.com
# Copyright (1999) Carnegie Mellon University, braam@cs.cmu.edu
# 
package Lento::Packet;
require Exporter;
use Lento::Debuggable;

@Lento::Packet::ISA = qw(Exporter Lento::Debuggable);
$Lento::Packet::level = -1000;
#use strict;

BEGIN {
    use Exporter   ();
    use vars       qw(@EXPORT %EXPORT_TAGS);
    @EXPORT      = qw(PKT_SIZE PKT_CTOKEN PKT_STOKEN PKT_IUID PKT_TYPE PKT_BODY
		      PKT_PARMS PKT_CONNECTION HDRSIZE);
    @EXPORT_OK   = qw();
    %EXPORT_TAGS = ( FIELDS => [ @EXPORT, @EXPORT_OK ] );
}

$Packet::REQ   = 0x1;  # always the 1st msg from the clnt -> srv
$Packet::REP   = 0x2;  # always srv -> clnt
$Packet::START = 0x3;  # begin I/O, srv -> clnt
$Packet::DAT   = 0x4;  # srv -> clnt  and clnt -> srv
$Packet::EOD   = 0x5;  # srv -> clnt  and clnt -> srv
$Packet::MSG   = 0x6;  # cln -> srv, always sent srv -> clnt packet arrived
$Packet::EOR   = 0x7;  # srv -> clnt

sub HDRSIZE        () { 20 } # size, ctoken, stoken, iuid, body_type
sub PKT_SIZE       () { 0 }
sub PKT_CTOKEN     () { 1 }
sub PKT_STOKEN     () { 2 }
sub PKT_IUID       () { 3 }
sub PKT_TYPE       () { 4 }
sub PKT_BODY       () { 5 }
sub PKT_PARMS      () { 6 }
sub PKT_CONNECTION () { 7 }

$Packet::str =  [undef, 'REQ', 'REP', 'START', 'DAT', 'EOD', 'MSG', 'EOR'];
my $packettype = { 'REQ' => $Packet::REQ, 'REP' => $Packet::REP, 
        'START' => $Packet::START, 'DAT' => $Packet::DAT, 
        'EOD' => $Packet::EOD, 'MSG' => $Packet::MSG, 'EOR' => $Packet::EOR };

sub new {
    my $class = $_[0];
    return bless( [ 0, 0, 0, 0, 0, '', [] ] , $class);
}

sub clone {
    my $oldpkt = shift;

    my $newpkt = new Lento::Packet;
    $newpkt->[PKT_CTOKEN] = $oldpkt->[PKT_CTOKEN];
    $newpkt->[PKT_STOKEN] = $oldpkt->[PKT_STOKEN];
    $newpkt->[PKT_IUID] = $oldpkt->[PKT_IUID];

    return $newpkt;
}

sub event_type {
    my $self = $_[0];

    return $Packet::str->[$self->[PKT_TYPE]];
}

sub getheader {
    my ($self, $buf, $offset) = @_;

    ($self->[PKT_SIZE], $self->[PKT_CTOKEN], $self->[PKT_STOKEN],
     $self->[PKT_IUID], $self->[PKT_TYPE]) = unpack "x${offset}N5", $buf;

#    DEBUG $self->printme();
}

# class method
sub init {
    my ($self, $type, $ctoken, $stoken, $iuid, $body) = @_;

    my $size = length($body);
    $self->[PKT_TYPE] = $type;
    $self->[PKT_SIZE] = $size;
    $self->[PKT_BODY] = $body;
    $self->[PKT_CTOKEN] = $ctoken;
    $self->[PKT_STOKEN] = $stoken;
    $self->[PKT_IUID] = $iuid;

    return $self;
}

sub packme {
    my $self = $_[0];
    my $size = length($self->[PKT_BODY]);

    return pack ("NNNNNa$size", $self->[PKT_SIZE], $self->[PKT_CTOKEN],
		 $self->[PKT_STOKEN], $self->[PKT_IUID], $self->[PKT_TYPE], 
		 $self->[PKT_BODY]);
}

sub len ($) {
    return $_[0]->[PKT_SIZE];
}

sub getbody ($$) {
    my ($self, $buf) = @_;

    die "Read error!" if $self->[PKT_SIZE] < length($buf);
    $self->[PKT_BODY] = substring($buf, HDRSIZE, $self->[PKT_SIZE]);
}

sub printme {
    my $self = $_[0];

    return("type: $self->[PKT_TYPE], size: $self->[PKT_SIZE], ctoken: $self->[PKT_CTOKEN], stoken: $self->[PKT_STOKEN], iuid: $self->[PKT_IUID]\n");
}

1;

__END__
