package Lento::RPCS::GetPermit;

use strict;
use Carp;

use POE;
use Lento::Connection;
use Lento::Debuggable;
use Lento::Packet;
use Lento::ReqDispatcher;
use Lento::Server;
use Lento::EventSupport; 
use Lento::Fileset;
use POSIX qw(:errno_h :fcntl_h);

@Lento::RPCS::GetPermit::ISA=qw (Debuggable);
$Lento::RPCS::GetPermit::level = 1;

sub GetPermit {
  my $fileset= shift || confess "no fileset";
  my $closure = shift || confess "No closure passed to GetPermit";
  my $session_level= shift;

  my $session = new POE::Session
      ([$fileset, $session_level], 
        $::dispatcher, ['connectme'],
        'Lento::EventSupport' , ['alias_set', 'alias_remove', '_stop', 
                                '_child', '_signal', 'AddWaiter'], 
        _start => sub {
                $_[HEAP]->{fileset}= $_[ARG0];
                $_[HEAP]->{session_level}= $_[ARG1];

                $_[HEAP]->{path}= $_[HEAP]->{fileset}->{mtpt};
                $_[HEAP]->{session_name} = $_[SESSION]->ID().
                            ": GetPermit ($_[HEAP]->{path}) $_[SESSION]";
                $_[KERNEL]->alias_set("GetPermit - $_[SESSION]");

                Lento::Debuggable::set_option
                    ($_[SESSION], trace => 1, default => 1);
           
                $_[HEAP]->{session_level} = LVL_FULL unless 
                    defined $_[HEAP]->{session_level};
                DEBUG("GetPermit::SESSION LVL set to $_[HEAP]->{session_level}\n");

                $_[HEAP]->{current_permitholder} = 
                    $_[HEAP]->{fileset}->{permitholder};
                $_[HEAP]->{current_permitholder} = "" 
                unless defined $_[HEAP]->{current_permitholder};
           
                LOG(">>>>> GetPermit RPC for fileset %s, permit currently at %s\n", 
                    $_[HEAP]->{fileset}->name(), $_[HEAP]->{current_permitholder});

            # We now call GetPermit without checking for current
                        # permitholder, so this might happen.
            if ($_[HEAP]->{current_permitholder} eq $::mysysid) {
                DEBUG " Already have permit!\n";
                        $_[KERNEL]->yield('done', 0, 0);
                return;
            }
       
                # if we are the server we should connect to the permit
                # holder, otherwise we connect to the server
                if ( $_[HEAP]->{fileset}->ishere() ) {
                if ($_[HEAP]->{current_permitholder} eq "") {
                    DEBUG " Fileset is here. Assuming I am permit holder\n";
                    DEBUG "taking permit\n";
                        $_[HEAP]->{fileset}->{permitholder} = $::mysysid;
                            $_[KERNEL]->yield('done', 0, 0);
                    return;
                }
                
                    #find the client
                    my $connection = Lento::Connection::find(
                                        $_[HEAP]->{current_permitholder});
                    if ( Lento::Connection::connected($connection)) {
                            # client still there: send request
                                        DEBUG " connected client sending req\n";
                            $_[HEAP]->{connection} = $connection;
                            $_[KERNEL]->yield('sendreq');
                    } else {
                        if ($_[HEAP]->{fileset}->{steal_permit}) { 
                        # We are allowed to steal the permit..
                            # tell the kernel to go on but do not 
                        # change fileset status
                                LOG(" Client gone, stealing permit\n");
                            LOG ">>>> Stealing permit from $_[HEAP]->{fileset}->{permitholder}\n";
                        DEBUG "taking permit\n";
                            $_[HEAP]->{fileset}->{permitholder}= $::mysysid;
                            $_[KERNEL]->yield('done', 0, 0);
                    } else {
                                                DEBUG " Disconnected client...raising got_error\n";
                        $_[KERNEL]->yield('got_error','getpermit', 
                                            ENOTCONN, 'no connection');
                    }
                    }
                } else {
                    # go to the server 
                    $_[KERNEL]->call($_[SESSION], 'connectme', $_[HEAP]->{fileset}, 
                                   $_[HEAP]->{session_level});
                }
        },
        got_connection => sub {
                my $fset = $_[HEAP]->{fileset};
                $_[HEAP]->{connection} = $fset->server()->{connection_obj};
                        # Make a call since we want to register for errors without 
                        # blocking during the event dispatch wait period.
                $_[KERNEL]->call($_[SESSION],'sendreq');
        },
        sendreq => sub {
            # We may get here on the server after acquiring the permit during
            # a delay in a got_error state.
            if ($_[HEAP]->{fileset}->{permitholder} eq $::mysysid) {
                DEBUG " Already have permit!\n";
                        $_[KERNEL]->yield('done', 0, 0);
                return;
            }
                        DEBUG "Now sending the request\n";
                $_[HEAP]->{connection}->sendpacket
                    ('REQ', ['Permit', $_[HEAP]->{fileset}->name()]);
        },
        REP => sub {
                my $packet = $_[ARG0];
                my $rc = shift @{$packet->[PKT_PARMS]};
            
            if ($rc) {
                    if ( $rc == EROFS ) {
                        LOG(">>>> ERROR : GetPermit returns EROFS for fileset %s. Set RO!\n".
                            $_[HEAP]->{fileset}->{name});
                        $_[HEAP]->{fileset}->{lento_read_only}= 0;
                        $_[HEAP]->{fileset}->{presto_read_only}= 1;
                    } else {
                    $!=$rc;
                            LOG("Unable to obtain permit: peer returned error $!\n");
                    $_[KERNEL]->yield('got_error','getpermit',$rc,'peer error');
                }
                } else {
                    DEBUG(">>>>  GetPermit REP: permit is now here ($::mysysid)\n");
                    DEBUG "taking permit\n";
                    $_[HEAP]->{fileset}->{permitholder} = $::mysysid;
            }

                $poe_kernel->yield('done', $rc, 0);
        },
        got_error => sub {
            my $errstr=$_[ARG0];
            my $errno=$_[ARG1];
            my $errmsg=$_[ARG2];

            $errno=ETIMEDOUT unless defined $errno and $errno != 0;

                # we cannot get a connection
                if ($_[HEAP]->{fileset}->{steal_permit}) { 
                # We are allowed to steal the permit..
                    # tell the kernel to go on but do not change fileset status
                    LOG ">>>> Stealing permit from $_[HEAP]->{fileset}->{permitholder}\n";
                DEBUG "taking permit\n";
                    $_[HEAP]->{fileset}->{permitholder}= $::mysysid;
                    $_[KERNEL]->yield('done', 0, 1);
             } else {
                # We are not allowed to steal the permit; we must reconnect
                    my $permitholder = $_[HEAP]->{fileset}->{permitholder};
                    $permitholder = "" unless defined $permitholder;

                # We might get a got_error right after getting the permit;
                if ($permitholder eq $::mysysid) {
                    DEBUG " Already have permit!\n";
                            $_[KERNEL]->yield('done', 0, 1);
                    return;
                }
       
                $_[KERNEL]->yield('done',$errno, 1);
             }
        },
        done => sub {
                $_[HEAP]->{connection}->put() 
                if ($_[HEAP]->{connection} and !$_[ARG1]);
                DEBUG(">>>   Calling WakeUpWaitList\n");
                Lento::EventSupport::WakeUpWaitList($_[ARG0]);
                $_[KERNEL]->alias_remove("GetPermit - $_[SESSION]");
        }
       );
  
  Lento::EventSupport::WaitForSession($session, $closure);
  
  return $session;
}

1;
