package Lento::ServerSemantics;
use strict;
use Lento::Debuggable;

@Lento::ServerSemantics::ISA = qw(Lento::Debuggable);
$Lento::ServerSemantics::level = -1000;

my $callback_timeout = 10;

sub new {
    my $class = shift || die "No class in new Callback";
    my $self = {};

    $self->{clients} = {};
    $self->{fsos} = {};
    $self->{filesets} = {};
    $self->{permits} = {};
    bless $self, $class;
}

#
# initialise an empty table of files for the client passed
#
sub add_client($$) {
    my $self = shift || die;
    my $client = shift || die;

    $self->{clients}->{$client}->{fsos} = {};
    $self->{clients}->{$client}->{permits} = {};
    $self->{clients}->{$client}->{filesets} = {};
}


#
# Register a callback promise to $client for ($fset, $file)
#
sub add_file($$$) {
    my $self = shift || die;
    my $client = shift || die;
    my $fset = shift || die;
    my $file = shift || die;

}

#
# Register a callback promise to $client for $fso 
#
sub add_fileset($$$) {
    my $self = shift || die;
    my $client = shift || die;
    my $fileset = shift || die;
    $self->{clients}->{$client}->{$fileset} = 'yes';
    $self->{filesets}->{$fileset}->{$client} = 'yes';
}

#
# client has disconnected: remove all callbacks for it
#
sub remove_client ($$) {
    my $self = shift || die;
    my $client = shift || die;
    my $fso;

    delete $self->{clients}->{$client};
    
    foreach $fso (keys %{$self->{fsos}}) {
	delete $self->{fsos}->{$fso}->{$client};
    }
}

#
# Send asynchronous revocation of a callback promise with a hint
#
sub revoke_callback ($$$) {
    my $self = shift || die;
    my $fso = shift || die;
    my $hint = shift;
    my $client;

    return unless exists $self->{fsos}->{$fso};

    foreach  $client ( keys %{$self->{fsos}->{$fso}} ) {
	$client->{CallbackService}->revoke($fso->{name}, $hint, 
					     [$client, $fso] );
    }

}

#
# Completion of the revocation
#
sub revoke_complete ($$$) {
    my $self = shift || die;
    my $cookie = shift || die;

    my ($client, $fso) = shift@$cookie;

    if (exists $self->{fsos}->{$fso}->{$client}) {
	delete $self->{fsos}->{$fso}->{$client};
    }
    if (exists $self->{clients}->{$client}->{$fso}) {
	delete $self->{clients}->{$client}->{$fso};
    }

}

#
# timeout of the revocation
# 
sub revoke_handle_timeout ($) {
    my $self = shift || die;
    my $client = shift;

    $self->remove_client($client) if $client;
}

1;
