# utilities for manipulating Lento stuff

package Lento::Utils;

use strict;
use Carp;
use Lento::Debuggable;
use File::stat;
use POSIX qw(:unistd_h :fcntl_h);
use Lento::KML::Rec;
use Lento::Psdev;
    
@Lento::Utils::ISA = qw(Lento::Debuggable);
$Lento::Utils::level = 1;

my $mntopt="-o loop,rw";
my $type='ext2';
my $izodir='/usr/src/intermezzo';
#$curwd = `pwd`;
#chdir('$izodir' . '/lento/tests');
my $filesize = 10254;
#my $filesize = $ENV{FSIZE} if defined $ENV{FSIZE};


sub openjournalf {
    my ($class, $mountpoint, $fsetname, $prestodev) = @_;

    $::psdev = Lento::Psdev->new($prestodev);
    $::psdev->open();

    print "setfsetroot: mtpt: $mountpoint, fileset: $fsetname\n";
    return $::psdev->setfsetroot($mountpoint, $fsetname, 1, 0x0);
}

sub closejournalf {
    my ($class, $mountpoint, $prestodev) = @_;

    $::psdev = Lento::Psdev->new($prestodev);
    $::psdev->open();

    print "clearfsetroot: mtpt: $mountpoint\n";
    return $::psdev->clearfsetroot($mountpoint);
}

sub cleanup {
    my ($class, $loop, $mntpt,$prestodev) = @_;

    $class->closejournalf($mntpt,$prestodev);
    `/bin/umount $mntpt`;
    unlink($loop);
}

sub idev_clean {
    my ($class,$mntpt) = @_;
    `/bin/umount $mntpt`;
}

sub loop_clean {
    my ($class, $loop) = @_;
    unlink($loop);
}

sub loop_init {
    my ($class,$loop,$mntpt,$fsetname) = @_;

    print "dd if=/dev/zero of=$loop bs=$filesize count=10000\n";
    `/bin/dd if=/dev/zero of=$loop bs=$filesize count=10000`;

    print "mke2fs -F $loop\n";
    `/sbin/mke2fs -F $loop`;

    print "mount -t $type $mntopt $loop $mntpt\n";
    `/bin/mount -t $type $mntopt $loop $mntpt`;

    print "making dirs\n";
    mkdir($mntpt,0755);
    mkdir($mntpt . '/.intermezzo',0755);
    mkdir($mntpt . '/.intermezzo/' . $fsetname,'0755');

    print "touching files\n";
    `touch $mntpt/.intermezzo/$fsetname/kml`;
    `touch $mntpt/.intermezzo/$fsetname/last_rcvd`;

    print "umount $mntpt\n";
    `/bin/umount $mntpt`;

    return 1;
}

sub idev_init {
    my ($class,$loop,$mntpt,$fsetname,$presto,$prestodev) = @_;

    `mount -t InterMezzo $mntopt,cache_type=$type,fileset=$fsetname,mtpt=$mntpt,prestodev=$prestodev $loop $mntpt`;

    `echo 4095 > /proc/sys/intermezzo/$presto/debug`;
    `echo 4095 > /proc/sys/intermezzo/$presto/trace`;
    `echo MARK.$$ >> /var/log/messages`;

    `echo 1 > /proc/sys/intermezzo/$presto/no_filter`;
    return 1;
}

