/*
 * KML REINT
 *
 * Copryright (C) 1996 Arthur Ma <arthur.ma@mountainviewdata.com>
 *
 * Copyright (C) 2000 Mountainview Data, Inc.
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/mmu_context.h>
#include <linux/intermezzo_fs.h>
#include <linux/intermezzo_kml.h>
#include <linux/intermezzo_psdev.h>
#include <linux/intermezzo_upcall.h>

static void kmlreint_pre_secure (struct kml_rec *rec);
static void kmlreint_post_secure (struct kml_rec *rec);

static void kmlreint_pre_secure (struct kml_rec *rec)
{
        if (current->fsuid != current->uid)
                CDEBUG (D_KML, "reint_kmlreint_pre_secure: cannot setfsuid\n");
        if (current->fsgid != current->gid)
                CDEBUG (D_KML, "reint_kmlreint_pre_secure: cannot setfsgid\n");
        current->fsuid = rec->rec_uid;
        current->fsgid = rec->rec_gid;
}

static void kmlreint_post_secure (struct kml_rec *rec)
{
        current->fsuid = current->uid; 
        current->fsgid = current->gid;
        /* current->egid = current->gid; */ 
        /* ????????????? */
}

static char *reint_makepath (char *mtpt, char *path)
{
        char *name;
        PRESTO_ALLOC(name, char *, strlen (mtpt) + strlen (path) + 1);
        if (name == NULL)
                return NULL;
        sprintf (name, "%s%s", mtpt, path);
        CDEBUG (D_KML, "=====reint_makepath::%s\n", name);
        return name;
}

static int reint_create (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        int     error;
        char   *name;
        mm_segment_t old_fs;

        ENTRY;
        kmlreint_pre_secure (rec);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;

        name = reint_makepath (mtpt, rec->rec_path);
        if (name == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        CDEBUG (D_KML, "=====REINT_CREATE::%s\n", name);
        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_create(name, rec->rec_mode, &info);
        set_fs(old_fs);
        PRESTO_FREE (name, strlen (name));

        kmlreint_post_secure (rec);

        EXIT;
        return error;
}


static int reint_open (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        return 0;
}

static int reint_mkdir (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        int     error;
        char   *name;
        mm_segment_t old_fs;

        ENTRY;
        kmlreint_pre_secure (rec);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;
        name = reint_makepath (mtpt, rec->rec_path);
        if (name == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_mkdir (name, rec->rec_mode, &info);
        set_fs(old_fs);
        PRESTO_FREE (name, strlen (name));
        kmlreint_post_secure (rec);

        EXIT;
        return error;
}

static int reint_rmdir (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        char *name;
        int error;
        mm_segment_t old_fs;
        int len = strlen (mtpt) + strlen (rec->rec_path) +
                  strlen (rec->rec_targetpath) + 3;

        ENTRY;
        kmlreint_pre_secure (rec);
        PRESTO_ALLOC(name, char *, len);
        if (name == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        sprintf (name, "%s%s/%s", mtpt , rec->rec_path, rec->rec_targetpath);
        CDEBUG (D_KML, "*************name=%s", name);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;

        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_rmdir (name, &info);
        set_fs(old_fs);

        PRESTO_FREE (name, strlen (name));
        kmlreint_post_secure (rec);
        EXIT;
        return error;
}

static int reint_link (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        int     error;
        mm_segment_t old_fs;
        char   *oldname, *newname;

        ENTRY;
        kmlreint_pre_secure (rec);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;

        oldname = reint_makepath (mtpt, rec->rec_path);
        if (oldname == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        newname = reint_makepath (mtpt, rec->rec_targetpath);
        if (newname == NULL)
        {
                PRESTO_FREE (oldname, strlen (oldname));
                kmlreint_post_secure (rec);
                return -ENOMEM;
        }
        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_link (oldname, newname, &info);
        set_fs(old_fs);
        PRESTO_FREE (oldname, strlen (oldname));
        PRESTO_FREE (newname, strlen (newname));
        kmlreint_post_secure (rec);
        EXIT;
        return error;
}

static int reint_unlink (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        int     error, len;
        mm_segment_t old_fs;
        char   *name;

        ENTRY;
        len = strlen (mtpt) + strlen (rec->rec_path) + 
                strlen (rec->rec_targetpath) + 3;       

        kmlreint_pre_secure (rec);
        PRESTO_ALLOC(name, char *, len);
        if (name == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        sprintf (name, "%s%s/%s", mtpt, rec->rec_path, rec->rec_targetpath);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;

        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_unlink (name, &info);
        set_fs(old_fs);
        PRESTO_FREE (name, strlen (name));
        kmlreint_post_secure (rec);

        EXIT;
        return error;
}

static int reint_symlink (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        int     error;
        mm_segment_t old_fs;
        char   *oldname, *newname;

        ENTRY;
        kmlreint_pre_secure (rec);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;

        oldname = reint_makepath (mtpt, rec->rec_path); 
        if (oldname == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        newname = rec->rec_targetpath;
        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_symlink (newname, oldname, &info);
        set_fs(old_fs);
        PRESTO_FREE (oldname, strlen (oldname));
        kmlreint_post_secure (rec);
        EXIT;
        return error;
}

static int reint_rename (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        int     error;
        mm_segment_t old_fs;
        char   *oldname, *newname;

        ENTRY;
        kmlreint_pre_secure (rec);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;

        oldname = reint_makepath (mtpt, rec->rec_path);
        if (oldname == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        newname = reint_makepath (mtpt, rec->rec_targetpath);
        if (newname == NULL)
        {
                PRESTO_FREE (oldname, strlen (oldname));
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_rename (oldname, newname, &info);
        set_fs(old_fs);
        PRESTO_FREE (oldname, strlen (oldname));
        PRESTO_FREE (newname, strlen (newname));
        kmlreint_post_secure (rec);

        EXIT;
        return error;
}

static int reint_setattr (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        struct iattr   attr;
        int     error;
        char   *name;
        mm_segment_t old_fs;

        ENTRY;
        kmlreint_pre_secure (rec);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;

        attr.ia_valid = rec->rec_valid;
        attr.ia_mode = rec->rec_mode;
        attr.ia_uid = rec->rec_uid;
        attr.ia_gid = rec->rec_gid;
        attr.ia_size = rec->rec_objattr.size;
        attr.ia_ctime = rec->rec_objattr.ctime;
        attr.ia_mtime = rec->rec_objattr.mtime;
        attr.ia_atime = rec->rec_objattr.ctime;
        attr.ia_attr_flags = rec->rec_flags;

        name = reint_makepath (mtpt, rec->rec_path);
        if (name == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_setattr (name, &attr, &info);
        set_fs(old_fs);
        PRESTO_FREE (name, strlen (name));
        kmlreint_post_secure (rec);
        EXIT;
        return error;
}

static int reint_mknod (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        struct  lento_vfs_context info;
        int     error;
        char   *name;
        mm_segment_t old_fs;

        ENTRY;
        kmlreint_pre_secure (rec);

        info.slot_offset = slot_offset;
        info.recno = rec->rec_suffix.recno;
        info.kml_offset = rec->rec_kml_offset;
        info.flags = rec->rec_flags;

        name = reint_makepath (mtpt, rec->rec_path);
        if (name == NULL)
        {
                kmlreint_post_secure (rec);
                EXIT;
                return -ENOMEM;
        }
        old_fs = get_fs();
        set_fs(get_ds());
        error = lento_mknod (name, rec->rec_mode, 
                MKDEV(rec->rec_major, rec->rec_minor), &info);
        set_fs(old_fs);
        PRESTO_FREE (name, strlen (name));
        kmlreint_post_secure (rec);
        EXIT;
        return error;
}

int kml_reint (char *mtpt, int slot_offset, struct kml_rec *rec)
{
        int error = 0;
        switch (rec->rec_head.opcode)
        {
                case KML_CREATE:
                        error = reint_create (mtpt, slot_offset, rec);
                        break;
                case KML_OPEN:
                        error = reint_open (mtpt, slot_offset, rec);
                        break;
                case KML_CLOSE:
                        /* error = reint_close (mtpt, slot_offset, rec);
                           force the system to return to lento */
                        error = 1;
                        break;
                case KML_MKDIR:
                        error = reint_mkdir (mtpt, slot_offset, rec);
                        break;
                case KML_RMDIR:
                        error = reint_rmdir (mtpt, slot_offset, rec);
                        break;
                case KML_UNLINK:
                        error = reint_unlink (mtpt, slot_offset, rec);
                        break;
                case KML_LINK:
                        error =  reint_link (mtpt, slot_offset, rec);
                        break;
                case KML_SYMLINK:
                        error = reint_symlink (mtpt, slot_offset, rec);
                        break;
                case KML_RENAME:
                        error = reint_rename (mtpt, slot_offset, rec);
                        break;
                case KML_SETATTR:
                        error =  reint_setattr (mtpt, slot_offset, rec);
                        break;
                case KML_MKNOD:
                        error = reint_mknod (mtpt, slot_offset, rec);
                        break;
                default:
                        CDEBUG (D_KML, "wrong opcode::%d\n", rec->rec_head.opcode);
                        return -EBADF;
        }
        if (error != 0)
                CDEBUG (D_KML, "KML_ERROR::error = %d\n", error);
        return error;
}

