#ifndef __PRESTO_JOURNAL_H
#define __PRESTO_JOURNAL_H


#include <linux/version.h>

#define PRESTO_KML_MAJOR_VERSION 0x00010000
#define PRESTO_KML_MINOR_VERSION 0x00000001

#define PRESTO_OP_NOOP          0
#define PRESTO_OP_CREATE        1
#define PRESTO_OP_MKDIR         2
#define PRESTO_OP_UNLINK        3
#define PRESTO_OP_RMDIR         4
#define PRESTO_OP_CLOSE         5
#define PRESTO_OP_SYMLINK       6
#define PRESTO_OP_RENAME        7
#define PRESTO_OP_SETATTR       8
#define PRESTO_OP_LINK          9
#define PRESTO_OP_OPEN          10
#define PRESTO_OP_MKNOD         11
#define PRESTO_OP_WRITE         12
#define PRESTO_OP_RELEASE       13
#define PRESTO_OP_TRUNC         14
#define PRESTO_OP_SETEXTATTR    15
#define PRESTO_OP_DELEXTATTR    16

#define PRESTO_LML_DONE     1 /* flag to get first write to do LML */


struct journal_prefix {
        u32 len;
        u32 version; 
        u32 pid;
        u32 uid;
        u32 fsuid;
        u32 fsgid;
        u32 opcode;
        u32 ngroups;
        u32 groups[0];
};

struct big_journal_prefix {
        u32 len;
        u32 version; 
        u32 pid;
        u32 uid;
        u32 fsuid;
        u32 fsgid;
        u32 opcode;
        u32 ngroups;
        u32 groups[NGROUPS_MAX];
};


struct journal_suffix {
        unsigned long prevrec;  /* offset of previous record for dentry */
        int recno;
        int time;
        int len;
};

#ifdef CONFIG_KREINT
enum kml_opcode {
        KML_CREATE = 1,
        KML_MKDIR,
        KML_UNLINK,
        KML_RMDIR,
        KML_CLOSE,
        KML_SYMLINK,
        KML_RENAME,
        KML_SETATTR,
        KML_LINK,
        KML_OPEN,
        KML_MKNOD
};

struct presto_attr
{
        int     pa_mode;
        int     pa_uid;
        int     pa_gid;
        char    *pa_op;
        char    *pa_path;
        int     pa_targetlen;
        int     pa_target;
        int     pa_pathlen;
        int     pa_major,
                pa_minor;
};

struct kml_attr {
        __u64   size,
                mtime,
                ctime;
};

struct kml_rec
{
        struct journal_prefix rec_head;

        /* major attribs of KML */
        char    *rec_op;
        int     rec_mode;
        int     rec_uid;
        int     rec_gid;
        int     rec_pathlen;
        char    *rec_path;
        int     rec_targetlen;
        char    *rec_targetpath;
        int     rec_major,
                rec_minor;

        /* minor attribs of KML */
        int     rec_offs;
        int     rec_valid;
        int     rec_generation;

        __u64   rec_ino;
        int     rec_kml_offset;
        int     rec_flags;

        struct  presto_version rec_old_parentv;
        struct  presto_version rec_new_parentv;
        struct  presto_version rec_new_objectv;
        struct  presto_version rec_old_objectv;

        struct presto_version rec_new_tgtv;
        struct presto_version rec_old_tgtv;
        struct kml_attr rec_objattr;

        struct journal_suffix rec_suffix;
};

extern void kml_print  (char *binrec, int binlen);
extern struct kml_rec * kml_upack (char *binbuf, int binlen);
extern int presto_reintbuf (char *volname, char *recbuf, int  reclen, int *newpos, int *count);
#endif 

#endif

