/* loop_discard.c
 *
 * Program to create "errors" on the loopback device in order to simulate
 * disk failure or crash while testing journalling support in presto/ext3.
 * It currently does this by interfacing with the (modified) loop device to
 * make it discard write and/or read requests when desired.
 *
 * Andreas Dilger <adilger@turbolinux.com>, Peter Braam <braam@turbolinux.com>
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdlib.h>

int main(int argc, char *argv[])
{
	int fd, err;
	long val;

	if (argc != 3) {
		fprintf(stderr, "usage: %s <device> <discard>\n", argv[0]);
		fprintf(stderr, "\tdiscard = 1 for writes, 2 for reads\n");
		return 1;
	}
	val = strtoul(argv[2], 0, 0);
	fd = open(argv[1], O_RDONLY);
	if (fd < 0) {
		perror("initial fd error");
		return 1;
	}
	err = ioctl(fd, 0x4C04, val);
	if (err) {
		perror("ioctl failed");
		return 1;
	}

	close(fd);
	return(err);
}

