/* Program to create "errors" on the loopback device in order to simulate
 * disk failure while testing journalling support in presto.  It currently
 * does this by interfacing with the (modified) loop device to make it
 * read-only when desired.
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdlib.h>

struct lento_vfs_context {
	unsigned int slot_offset;
	unsigned int recno;
	unsigned int kml_offset;
	unsigned int flags;
};

struct lento_input_iopen {
	char *name;
	unsigned int dummy;
	unsigned long long ino;
	unsigned int generation;
	int flags;
	int fd;
};

struct lento_input_close {
	unsigned int fd;
	struct lento_vfs_context info;
};

#define PRESTO_VFS_IOPEN _IOWR('p', 0x1b, struct lento_input_iopen *)
#define PRESTO_VFS_CLOSE _IOW ('p', 0x1a, struct lento_input_close *)
#define BUFSIZE 4096
int main(int argc, char *argv[])
{
	int fd, err;
	char buf[BUFSIZE];
	struct lento_input_iopen input_iopen;
	struct lento_input_close input_close;
	int ret;

	if (argc != 4) {
		fprintf(stderr, "usage: %s <filename> <ino> <generation>\n", argv[0]);
		return 1;
	}
	fd = open("/dev/intermezzo0", O_RDONLY);
	if (fd < 0) {
		perror("initial fd error");
		return 1;
	}
	input_iopen.name = argv[1];
	input_iopen.ino = strtoul(argv[2], NULL, 0);
	input_iopen.generation = strtoul(argv[3], NULL, 0);
	input_iopen.flags = O_RDWR;
	err = ioctl(fd, PRESTO_VFS_IOPEN, &input_iopen);
	printf("we got fd %d back\n", input_iopen.fd);
	if (err) {
		perror("ioctl failed");
		return 1;
	}

	ret = read(input_iopen.fd, buf, BUFSIZE);
	err = write(fileno(stdout), buf, ret);
	printf ("\n****\nread %d bytes at 0 from fd %d\n", ret, input_iopen.fd);
	if (ret < 0)
		perror("read error");

#if 0
	err = write(input_iopen.fd, buf, ret);
	printf ("write %d bytes at %d to fd %d\n", err, ret, input_iopen.fd);
	if (err < 0)
		perror("write error");
#endif
	input_close.fd = input_iopen.fd;
	err = ioctl(fd, PRESTO_VFS_CLOSE, &input_close);
	if (err < 0)
		perror("close error");
	close(fd);
	return(err);
}

