/* Program to time the creation of files under presto. */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/time.h>


struct lento_vfs_context {
	int slot_offset;
	int recno;
	int kml_offset;
	int flags;
};

struct lento_input_mode {
	char *name;
	int mode;
	struct lento_vfs_context info;
};

#define PRESTO_VFS_CREATE _IOW('p', 0x11, struct lento_input_mode *)

int main(int argc, char *argv[])
{
	int fd;
	struct lento_input_mode input;
	struct timeval start, stop;
	char name[4096];
	double elapsed;
	int num;
	int i;

	if (argc < 3) {
		fprintf(stderr, "usage: %s <filename> <num>\n", argv[0]);
		return 1;
	}
	num = atol(argv[2]);


	fd = open("/dev/intermezzo0", O_RDONLY);
	if (fd < 0) {
		perror("initial fd error");
		return 1;
	}
	input.name = name;
	input.mode = 0644;
	input.info.slot_offset = 0;
	input.info.recno = 4711;
	input.info.kml_offset = 0;
	input.info.flags = 3;
	gettimeofday(&start, NULL);
	for (i = 0; i < num; i++) {
		snprintf(name, 4096, "/izo0/%s%05d", argv[1], i);
		/* printf("create %s\n", name); */
		if (ioctl(fd, PRESTO_VFS_CREATE, &input)) {
			perror("ioctl failed");
			break;
		}
	}
	gettimeofday(&stop, NULL);
	elapsed = (double)(stop.tv_sec - start.tv_sec) +
	       ((double)(stop.tv_usec - start.tv_usec) / 1000000);
	printf("start time %d.%06d stop time %d.%06d\n",
	       start.tv_sec, start.tv_usec, stop.tv_sec, stop.tv_usec);
	printf("ext3 IOCTL time %f (per op %f)\n", elapsed, elapsed / num);
	close(fd);

	gettimeofday(&start, NULL);
	for (i = 0; i < num; i++) {
		int newfd;

		snprintf(name, 4096, "/izo1/%s%05d", argv[1], i);
		newfd = open(name, O_RDONLY|O_CREAT, 0644);
		if (newfd < 0) {
			perror("open failed");
			break;
		}

		if (close(newfd) < 0) {
			perror("close error");
			break;
		}
	}
	gettimeofday(&stop, NULL);
	elapsed = (double)(stop.tv_sec - start.tv_sec) +
	       ((double)(stop.tv_usec - start.tv_usec) / 1000000);
	printf("start time %d.%06d stop time %d.%06d\n",
	       start.tv_sec, start.tv_usec, stop.tv_sec, stop.tv_usec);
	printf("ext3 time %f (per op %f)\n", elapsed, elapsed / num);

	gettimeofday(&start, NULL);
	for (i = 0; i < num; i++) {
		int newfd;

		snprintf(name, 4096, "/izo2/%s%05d", argv[1], i);
		newfd = open(name, O_RDONLY|O_CREAT, 0644);
		if (newfd < 0) {
			perror("open failed");
			break;
		}

		if (close(newfd) < 0) {
			perror("close error");
			break;
		}
	}
	gettimeofday(&stop, NULL);
	elapsed = (double)(stop.tv_sec - start.tv_sec) +
	       ((double)(stop.tv_usec - start.tv_usec) / 1000000);
	printf("start time %d.%06d stop time %d.%06d\n",
	       start.tv_sec, start.tv_usec, stop.tv_sec, stop.tv_usec);
	printf("ext2 time %f (per op %f)\n", elapsed, elapsed / num);

	return 0;
}

